import pika


class RabbitMQClient(object):
    def __init__(self, config):
        self.queue = config.rabbitmq_queue
        self.exchange = config.rabbitmq_exchange
        self.host = config.rabbitmq_host
        self.type = None

    def __enter__(self):
        self.connection = pika.BlockingConnection(pika.ConnectionParameters(self.host))
        self.channel = self.connection.channel()
        return self

    def __exit__(self, type, value, traceback):
        self.connection.close()


class RabbitMQPublisher(RabbitMQClient):
    def push_message(self, body):
        if not self.type:
            self.set_publisher()
        assert self.type == 'publisher'
        self.channel.basic_publish(exchange=self.exchange, routing_key=self.queue, body=body)

    def set_publisher(self):
        self.channel.queue_declare(queue=self.queue)
        self.type = 'publisher'


class RabbitMQConsumer(RabbitMQClient):
    def consume(self, callback):
        self.channel.basic_consume(callback, queue=self.queue, no_ack=True)
        self.channel.start_consuming()




import datetime


class Message(object):
    def __init__(self, service, metric, time=None):
        self.service = service
        self.time = time or datetime.datetime.now()
        self.metric = metric

    def to_dict(self):
        return dict(service=self.service,
                    time=self.time.strftime('%Y-%m-%d %H:%M:%S%z'),
                    metric=self.metric)



from ..rabbitmq.message import Message


class MetricsCollector(object):
    def __init__(self):
        self.collectors = []

    def register(self, collector):
        self.collectors.append(collector)

    def collect_metrics(self):
        return [message for collector in self.collectors for message in collector.collect_metrics()]


class BaseCollector(object):
    def collect_metrics(self) -> list:
        raise NotImplementedError()



import click

from cassandra.cluster import Cluster
from cassandra.cluster import NoHostAvailable

from ..consumer.consumer import ConsumerClient
from ..rabbitmq.message import Message


class CassandraClient(ConsumerClient):

    def __init__(self, config):
        self.cluster = Cluster(
            contact_points=config.cassandra_endpoints.split(','),
            port=9042,
        )
        self.session = None

    def connect(self):
        if self.session and not self.session.is_shutdown:
            click.echo("already connected")

        try:
            self.session = self.cluster.connect(keyspace='metrics')
        except NoHostAvailable as e:
            click.echo(
                "\nError....couldn't connect to Cassandra"
            )
            raise e

    def disconnect(self):
        self.cluster.shutdown()

    def __del__(self):
        self.disconnect()

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, type, value, traceback):
        self.disconnect()

    def recieve_message(self, message):
        for hostname, metrics in message.items():
            messages = [Message(**metric) for metric in metrics]

            for message in messages:
                bucket_time = ':'.join(message.time.split(':')[:-1])[:-1]+'0'
                self.session.execute(
                    """
                    INSERT INTO metrics.raw (host, service, bucket_time, time, metric)
                    VALUES (%s, %s, %s, %s, %s)
                    """, (hostname, message.service, bucket_time, message.time, message.metric)
                )



import psutil

from ..collector.collector import BaseCollector
from ..rabbitmq.message import Message


class MemoryMetricsCollector(BaseCollector):
    def collect_metrics(self):
        return [
            Message('memory/' + k, v)
            for k, v in psutil.virtual_memory()._asdict().items()
        ]



import datetime

import psutil

from ..rabbitmq.message import Message
from ..collector.collector import BaseCollector


class NetworkMetricsCollector(BaseCollector):
    def __init__(self):
        self.last_bytes_sent = 0
        self.last_bytes_recv = 0
        self.last_packets_sent = 0
        self.last_packets_recv = 0
        self.last_errin = 0
        self.last_errout = 0
        self.last_dropin = 0
        self.last_dropout = 0
        self.last_collect_time = datetime.datetime.now()

    def collect_metrics(self):
        network_stats = psutil.net_io_counters()
        collect_time = datetime.datetime.now()
        collect_interval = (collect_time - self.last_collect_time).seconds
        self.last_collect_time = collect_time
        collect_interval = 1 if collect_interval < 1 else collect_interval
        messages = [
            Message('net/bytes_sent/sec', (network_stats.bytes_sent - self.last_bytes_sent) / collect_interval),
            Message('net/bytes_recv/sec', (network_stats.bytes_recv - self.last_bytes_recv) / collect_interval),
            Message('net/packets_sent/sec', (network_stats.packets_sent - self.last_packets_sent) / collect_interval),
            Message('net/packets_recv/sec', (network_stats.packets_recv - self.last_packets_recv) / collect_interval),
            Message('net/errin/sec', (network_stats.errin - self.last_errin) / collect_interval),
            Message('net/errout/sec', (network_stats.errout - self.last_errout) / collect_interval),
            Message('net/dropin/sec', (network_stats.dropin - self.last_dropin) / collect_interval),
            Message('net/dropout/sec', (network_stats.dropout - self.last_dropout) / collect_interval)
        ]
        self.last_bytes_sent = network_stats.bytes_sent
        self.last_bytes_recv = network_stats.bytes_recv
        self.last_packets_sent = network_stats.packets_sent
        self.last_packets_recv = network_stats.packets_recv
        self.last_errin = network_stats.errin
        self.last_errout = network_stats.errout
        self.last_dropin = network_stats.dropin
        self.last_dropout = network_stats.dropout

        return messages



import psutil

from ..collector.collector import BaseCollector
from ..rabbitmq.message import Message


class CPUMetricsCollector(BaseCollector):
    def collect_metrics(self):
        cpu_times = psutil.cpu_times()
        return [
            Message(service='cpu/percent', metric=psutil.cpu_percent()),
            Message(service='cpu/times_user', metric=cpu_times.user),
            Message(service='cpu/times_nice', metric=cpu_times.nice),
            Message(service='cpu/times_system', metric=cpu_times.system),
            Message(service='cpu/times_idle', metric=cpu_times.idle),
            Message(service='cpu/times_iowait', metric=cpu_times.iowait),
            Message(service='cpu/times_irq', metric=cpu_times.irq),
            Message(service='cpu/times_softirq', metric=cpu_times.softirq),
            Message(service='cpu/times_steal', metric=cpu_times.steal),
            Message(service='cpu/times_guest', metric=cpu_times.guest),
            Message(service='cpu/times_guest_nice', metric=cpu_times.guest_nice)
            ]



from riemann_client.transport import TCPTransport
from riemann_client.client import QueuedClient

from ..rabbitmq.message import Message
from ..consumer.consumer import ConsumerClient


class RiemannClient(ConsumerClient):
    def __init__(self, config):
        self.client = QueuedClient(TCPTransport(config.riemann_host, config.riemann_port))
        self.client.transport.connect()

    def recieve_message(self, message):
        for hostname, metrics in message.items():
            for metric in metrics:
                message = Message(**metric)
                self.client.event(service=message.service, metric_f=message.metric, host=hostname)

            self.client.flush()



import logging


class Logger(object):
    def __init__(self):
        self.logger = logging.getLogger('chestymetrics')

    def info(self, msg):
        self.logger.info(msg=msg)

    def warn(self, msg):
        self.logger.warning(msg=msg)

    def error(self, msg):
        self.logger.error(msg=msg)



import json

import click
import socket

import time

from .lib.collector.memory import MemoryMetricsCollector
from .lib.collector.disk import DiskMetricsCollector
from .lib.collector.network import NetworkMetricsCollector
from .lib.riemann.client import RiemannClient
from .lib.cassandra.client import CassandraClient
from .lib.consumer.consumer import Consumer
from .lib.collector.cpu import CPUMetricsCollector
from .lib.collector.collector import MetricsCollector
from .lib.logger.logger import Logger
from .lib.config.config import Config
from .lib.rabbitmq.client import RabbitMQConsumer, RabbitMQPublisher


@click.group()
def cli():
    pass


@cli.command()
@click.option('--rabbitmq-host', default='192.168.1.2', type=click.types.STRING)
@click.option('--rabbitmq-port', default=5672, type=click.types.INT)
@click.option('--rabbitmq-queue', default='chestymetrics', type=click.types.STRING)
@click.option('--rabbitmq-exchange', default='', type=click.types.STRING)
@click.option('--interval', default=60.0)
def reporter(**kwargs):
    config = Config(**kwargs)
    logger = Logger()
    logger.info('Starting chestymetrics...')
    logger.info('connecting to RabbitMQ...')
    with RabbitMQPublisher(config) as rabbitmq_client:
        logger.info('connected.')
        hostname = socket.gethostname()
        metrics_collector = MetricsCollector()
        metrics_collector.register(CPUMetricsCollector())
        metrics_collector.register(NetworkMetricsCollector())
        metrics_collector.register(DiskMetricsCollector())
        metrics_collector.register(MemoryMetricsCollector())

        starttime=time.time()
        while True:
            rabbitmq_client.push_message(
                json.dumps({hostname: [message.to_dict() for message in metrics_collector.collect_metrics()]})
            )
            time.sleep(kwargs.get('interval') - ((time.time() - starttime) % kwargs.get('interval')))


@cli.command()
@click.option('--rabbitmq-host', default='192.168.1.2', type=click.types.STRING)
@click.option('--rabbitmq-port', default=5672, type=click.types.INT)
@click.option('--rabbitmq-queue', default='chestymetrics', type=click.types.STRING)
@click.option('--rabbitmq-exchange', default='', type=click.types.STRING)
@click.option('--cassandra-endpoints', default='192.168.1.51,192.168.1.52,192.168.1.53', type=click.types.STRING)
@click.option('--riemann-host', default='192.168.1.2', type=click.types.STRING)
@click.option('--riemann-port', default=5555, type=click.types.INT)
def consumer(**kwargs):
    config = Config(**kwargs)
    logger = Logger()
    logger.info('starting chestymetrics consumer...')
    logger.info('connecting to RabbitMQ...')
    consumer = Consumer()
    with RabbitMQConsumer(config) as rabbitmq_client:
        with CassandraClient(config) as cassandra_client:
            consumer.register(cassandra_client)
            consumer.register(RiemannClient(config))
            logger.info('connected')
            rabbitmq_client.consume(consumer.callback)

if __name__ == '__main__':
    cli()



from setuptools import setup, find_packages

setup(
    name='Chestymetrics',
    version='0.0.1-beta',
    packages=find_packages(),
    url='http://www.google.com',
    license='GPLv3',
    author='max',
    author_email='max@max.com',
    description='foo',
    requires=[
        'pika',
        'click',
        'psutil',
        'cassandra_driver',
        'riemann_client'
    ],
    entry_points=
    '''
    [console_scripts]
    chestymetrics=chestymetrics.__init__:cli
    '''
)



import datetime

import psutil

from ..rabbitmq.message import Message
from ..collector.collector import BaseCollector


class DiskMetricsCollector(BaseCollector):
    def __init__(self):
        self.partitions = psutil.disk_partitions()
        self.last_io_counters = None
        self.last_collect_time = datetime.datetime.now()

    def collect_metrics(self):
        messages = []
        io_counters = psutil.disk_io_counters()._asdict()
        collect_time = datetime.datetime.now()
        collect_interval = (collect_time - self.last_collect_time).seconds
        self.last_collect_time = collect_time
        for part in self.partitions:
            disk_usage = psutil.disk_usage(part.device)
            messages.extend(
                [
                    Message(part.mountpoint + '/total_bytes', disk_usage.total),
                    Message(part.mountpoint + '/used_bytes', disk_usage.used),
                    Message(part.mountpoint + '/free_bytes', disk_usage.free),
                    Message(part.mountpoint + '/percent_used', disk_usage.percent)
                ]
            )
        if self.last_io_counters is not None:
            messages.extend(
                [
                    Message('disk/' + k + '/sec', (v - self.last_io_counters[k]) / collect_interval)
                    for k, v in psutil.disk_io_counters()._asdict().items()
                ]
            )
        self.last_io_counters = io_counters
        return messages



class Config(object):
    def __init__(self, **kwargs):
        self.rabbitmq_host = kwargs.get('rabbitmq_host')
        self.rabbitmq_port = kwargs.get('rabbitmq_port')
        self.rabbitmq_queue = kwargs.get('rabbitmq_queue')
        self.rabbitmq_exchange = kwargs.get('rabbitmq_exchange')
        self.cassandra_endpoints = kwargs.get('cassandra_endpoints')
        self.riemann_host = kwargs.get('riemann_host')
        self.riemann_port = kwargs.get('riemann_port')



import json


class Consumer(object):
    def __init__(self):
        self.listeners = []

    def register(self, listener):
        self.listeners.append(listener)

    def callback(self, ch, method, properties, body):
        message = json.loads(body)
        for listener in self.listeners:
            listener.recieve_message(message)


class ConsumerClient(object):
    """
    template class for clients that consume messages to extend
    """
    def recieve_message(self, message):
        raise NotImplementedError()
